# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import Hashable, Iterable
from dataclasses import dataclass
from typing import Any
from uuid import uuid4

import networkx as nx

from .node_values import IndexName, IndexValue, NodeValues


def _get_unique_sink(graph: nx.DiGraph) -> Hashable:
    sink_nodes = [node for node in graph.nodes if graph.out_degree(node) == 0]
    if len(sink_nodes) != 1:
        raise ValueError('Graph must have exactly one sink node')
    return sink_nodes[0]


def _get_new_node_name(graph: nx.DiGraph) -> str:
    while True:
        name = str(uuid4())
        if name not in graph:
            return name


def _remove_ancestors(graph: nx.DiGraph, node: Hashable) -> nx.DiGraph:
    graph_without_node = graph.copy()
    graph_without_node.remove_node(node)
    ancestors = nx.ancestors(graph, node)
    # Considering the graph we obtain by removing `node`, we need to consider the
    # descendants of each ancestor. If an ancestor has descendants that are not
    # removal candidates, we should not remove the ancestor.
    to_remove = [
        ancestor
        for ancestor in ancestors
        if nx.descendants(graph_without_node, ancestor).issubset(ancestors)
    ]
    graph = graph.copy()
    graph.remove_nodes_from(to_remove)
    graph.remove_edges_from(list(graph.in_edges(node)))
    return graph


@dataclass(frozen=True, slots=True)
class IndexValues:
    """
    Index values used as part of :py:class:`NodeName`.

    Conceptually, this is a mapping from index names to index values.
    """

    axes: tuple[IndexName, ...]
    values: tuple[IndexValue, ...]

    def to_tuple(self) -> tuple[tuple[IndexName, IndexValue], ...]:
        return tuple(zip(self.axes, self.values, strict=True))

    def merge_index(self, other: IndexValues) -> IndexValues:
        return IndexValues(
            axes=other.axes + self.axes, values=other.values + self.values
        )

    def __str__(self) -> str:
        return ', '.join(
            f'{name}={value}'
            for name, value in zip(self.axes, self.values, strict=True)
        )

    def __len__(self) -> int:
        return len(self.axes)


@dataclass(frozen=True, slots=True)
class NodeName:
    """Node name with indices used for mapped nodes when converting to NetworkX."""

    name: Hashable
    index: IndexValues

    def merge_index(self, other: IndexValues) -> NodeName:
        return NodeName(name=self.name, index=self.index.merge_index(other))

    def __str__(self) -> str:
        return f'{self.name}({self.index})'


@dataclass(frozen=True, slots=True)
class MappedNode:
    """
    Key for a node in :py:class:`Graph` representing a collection of "mapped" nodes.
    """

    name: Hashable
    indices: tuple[IndexName, ...]


def _node_with_indices(node: Hashable, indices: tuple[IndexName, ...]) -> MappedNode:
    if isinstance(node, MappedNode):
        return MappedNode(name=node.name, indices=indices + node.indices)
    return MappedNode(name=node, indices=indices)


def _node_indices(node: Hashable) -> tuple[IndexName, ...]:
    if isinstance(node, MappedNode):
        return node.indices
    return ()


def _find_successors(
    graph: nx.DiGraph, *, root_nodes: tuple[Hashable]
) -> set[Hashable]:
    successors = set()
    for root in root_nodes:
        if graph.in_degree(root) > 0:
            raise ValueError(f"Mapped node '{root}' is not a source node")
        successors.update(nx.descendants(graph, source=root) | {root})
    return successors


def _rename_successors(
    graph: nx.DiGraph, *, successors: Iterable[Hashable], index: IndexValues
) -> nx.DiGraph:
    """Replace 'node' and all its successors with (node, suffix), and update all edges
    accordingly."""
    renamed_nodes = {
        node: (
            node.merge_index(index)
            if isinstance(node, NodeName)
            else NodeName(name=node, index=index)
        )
        for node in successors
    }
    return nx.relabel_nodes(graph, renamed_nodes, copy=True)


class PositionalIndexer:
    """
    Helper class to allow slicing a named dim of a graph using positional indexing.
    """

    def __init__(self, graph: Graph, index_name: IndexName):
        self.graph = graph
        self.index_name = index_name

    def __getitem__(self, key: int | slice) -> Graph:
        # Supporting single indices may be conceptually ill-defined if the index
        # `reduce` was applied to the graph, so we might never support this.
        if isinstance(key, int):
            raise NotImplementedError('Only slices are supported')
        node_values = NodeValues(
            {
                name: (
                    col.loc({self.index_name: key})
                    if self.index_name in col.index_names
                    else col
                )
                for name, col in self.graph._node_values.items()
            }
        )
        return Graph(self.graph.graph, node_values=node_values)


MappingToArrayLike = Any  # dict[str, Numpy|DataArray], DataFrame, etc.


class Graph:
    """
    A Cyclebane graph is a directed acyclic graph with additional array-like structure.

    The array-like structure selectively affects nodes in the graph by associating
    source nodes with an array-like object. These source node and all their descendants
    thus gain an additional index (or "dimension").

    Nomenclature:

    - Index: As in Pandas, and index is a sequence of values that label an axis.
    - Index-value: A single value in an index.
    - Index-name: The name of an index.


    Notes
    -----
    The current implementation is not complete, there is a number of things to
    improve:
    - Overall, I would like to reduce the array-handling code and transparently forward
      to the slicing code of the underlying array-like object (Pandas, NumPy, Xarray,
      Scipp). Basically, we would like to use the slicing methods of the underlying
      object. This may not be trivial, since we might mix different types of array-like
      objects at nodes with multiple predecessors.
    """

    def __init__(self, graph: nx.DiGraph, *, node_values: NodeValues | None = None):
        """
        Initialize a graph from a directed NetworkX graph.

        Parameters
        ----------
        graph:
            The directed graph representing the data flow.
        node_values:
            A mapping from source node names to array-like objects. The implementation
            assumes that the graph has been setup correctly. Do not use this argument
            unless you know what you are doing.
        """
        self.graph = graph
        self._node_values = node_values or NodeValues({})

    def copy(self) -> Graph:
        return Graph(self.graph.copy(), node_values=self._node_values)

    @property
    def index_names(self) -> tuple[IndexName, ...]:
        """Names of the indices (dimensions) of the graph."""
        return tuple(self.indices)

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        """Names and values of the indices of the graph."""
        return self._node_values.indices

    def map(self, node_values: MappingToArrayLike) -> Graph:
        """
        Map the graph over the given values by associating source nodes with values.

        All successors of the mapped source nodes are replaced with new nodes, one for
        each index value. The value is set as an attribute on the new source nodes
        (but not their successors).

        Parameters
        ----------
        node_values:
            A mapping from source node names to array-like objects. The source nodes
            are the roots of the branches to be mapped. The array-like objects must
            support slicing, e.g., NumPy arrays, Xarray DataArrays, Pandas DataFrames,
            etc.
        """
        new_values = NodeValues.from_mapping(
            node_values, axis_zero=len(self.index_names)
        )

        # Make sure root nodes exist in graph, add them if not. This choice allows for
        # mapping, e.g., with multiple columns from a DataFrame, representing labels
        # used later for groupby operations.
        graph = self.graph.copy()
        graph.add_nodes_from(new_values)

        successors = _find_successors(graph, root_nodes=new_values)
        name_mapping: dict[Hashable, MappedNode] = {}
        for node in successors:
            name_mapping[node] = _node_with_indices(node, tuple(new_values.indices))

        return Graph(
            nx.relabel_nodes(graph, name_mapping),
            node_values=self._node_values.merge(new_values),
        )

    def reduce(
        self,
        key: None | Hashable = None,
        *,
        index: None | Hashable = None,
        axis: None | int = None,
        name: None | Hashable = None,
        attrs: None | dict[str, Any] = None,
    ) -> Graph:
        """
        Reduce over the given index or axis previously created with :py:meth:`map`.

        If neither index nor axis is given, all axes are reduced.

        Parameters
        ----------
        key:
            The name of the source node to reduce. This is the original name prior to
            mapping. If not given, tries to find a unique sink node.
        index:
            The name of the index to reduce over. Only one of index and axis can be
            given.
        axis:
            Integer axis index to reduce over. Only one of index and axis can be given.
        name:
            The name of the new node. If not given, a unique name is generated.
        attrs:
            Attributes to set on the new node(s).
        """
        key = key or _get_unique_sink(self.graph)
        name = name or _get_new_node_name(self.graph)

        attrs = attrs or {}
        if index is not None and axis is not None:
            raise ValueError('Only one of index and axis can be given')
        key = self._from_orig_key(key)
        indices = _node_indices(key)
        if index is not None and index not in indices:
            raise ValueError(f"Node '{key}' does not have index '{index}'.")
        # TODO We can support indexing from the back in the future.
        if axis is not None and (axis < 0 or axis >= len(indices)):
            raise ValueError(f"Node '{key}' does not have axis '{axis}'.")
        if index is not None:
            new_index = tuple(value for value in indices if value != index)
        elif axis is not None:
            # TODO Should axis refer to axes of graph, or the node?
            new_index = tuple(value for i, value in enumerate(indices) if i != axis)
        else:
            new_index = None
        if name in self.graph:
            raise ValueError(f"Node '{name}' already exists in the graph.")

        graph = self.graph.copy()
        name = MappedNode(name=name, indices=new_index) if new_index else name
        graph.add_node(name, **attrs)
        graph.add_edge(key, name)

        return Graph(graph, node_values=self._node_values)

    def _from_orig_key(self, key: Hashable) -> Hashable:
        # Graph.map relabels nodes to include index names, which can be inconvenient
        # for the user. Is this convenience of finding the node by its original name
        # worth the complexity and a good idea?
        if key not in self.graph:
            matches = [
                node
                for node in self.graph.nodes
                if isinstance(node, MappedNode) and node.name == key
            ]
            if len(matches) == 0:
                raise KeyError(f"Node '{key}' does not exist in the graph.")
            if len(matches) > 1:
                raise KeyError(f"Node '{key}' is ambiguous. Found {matches}.")
            return matches[0]
        return key

    def by_position(self, index_name: IndexName) -> PositionalIndexer:
        return PositionalIndexer(self, index_name)

    def to_networkx(self, value_attr: str = 'value') -> nx.DiGraph:
        """
        Convert to a NetworkX graph, spelling out the internal array structures as
        explicit nodes.

        Parameters
        ----------
        value_attr:
            The name of the attribute on nodes that holds the array-like object.
        """
        graph = self.graph
        for index_name, index in reversed(self.indices.items()):
            # Find all nodes with this index
            nodes = [
                node
                for node in graph.nodes()
                if index_name
                in _node_indices(node.name if isinstance(node, NodeName) else node)
            ]
            # Make a copy for each index value
            graphs = [
                _rename_successors(
                    graph, successors=nodes, index=IndexValues((index_name,), (i,))
                )
                for i in index
            ]
            graph = nx.compose_all(graphs)
        # Replace all MappingNodes with their name
        new_names = {
            node: NodeName(node.name.name, node.index)
            for node in graph
            if isinstance(node, NodeName) and isinstance(node.name, MappedNode)
        }
        graph = nx.relabel_nodes(graph, new_names)

        # Get values using previously stored index values
        for node in graph.nodes:
            if (
                isinstance(node, NodeName)
                and (node_values := self._node_values.get(node.name)) is not None
            ):
                graph.nodes[node][value_attr] = node_values.sel(node.index.to_tuple())

        return graph

    def __getitem__(self, key: Hashable | slice) -> Graph:
        """
        Get the branch of the graph rooted at the given node.

        The branch is a subgraph containing the given node and all its ancestors.
        Think of this like a Git branch, where the given node is the head of the branch.
        """
        if isinstance(key, slice):
            raise NotImplementedError('Only single nodes are supported ')
        key = self._from_orig_key(key)
        ancestors = nx.ancestors(self.graph, key)
        ancestors.add(key)
        # Drop all node values that are not in the branch
        mapped = {a.name for a in ancestors if isinstance(a, MappedNode)}
        keep_values = [key for key in self._node_values.keys() if key in mapped]
        return Graph(
            self.graph.subgraph(ancestors),
            node_values=self._node_values.get_columns(keep_values),
        )

    def __delitem__(self, key: Hashable | slice) -> None:
        """
        Delete the branch of the graph rooted at the given node.
        """
        if isinstance(key, slice):
            raise NotImplementedError('Only single nodes are supported ')
        key = self._from_orig_key(key)
        if isinstance(key, MappedNode):
            # Not clear what to do in this case, as it would leave a lot of MappedNodes
            # without a source that could provide data.
            raise ValueError('Cannot delete mapped node.')
        graph = _remove_ancestors(self.graph, key)
        graph.nodes[key].clear()
        mapped = {node.name for node in graph if isinstance(node, MappedNode)}
        keep_values = [key for key in self._node_values.keys() if key in mapped]
        self._node_values = self._node_values.get_columns(keep_values)
        self.graph = graph

    def __setitem__(self, branch: Hashable | slice, other: Graph) -> None:
        """
        Set a new branch in place of the given branch.

        The new branch must have a unique sink node. The branch at `branch` is replaced
        with the new branch. The indices and node values are updated accordingly. The
        edges to successors of the old branch are connected to the sink of the new
        branch.
        """
        if isinstance(branch, slice):
            raise NotImplementedError('Setting slice not supported yet.')
        if not isinstance(other, Graph):
            raise TypeError(f'Expected {Graph}, got {type(other)}')
        new_branch = other.graph
        sink = _get_unique_sink(new_branch)
        # In the future, we could support this if BOTH sink and branch are MappedNodes
        # with identical indices.
        if isinstance(sink, MappedNode) or isinstance(branch, MappedNode):
            raise NotImplementedError('Mapped nodes not supported yet in __setitem__')
        new_branch = nx.relabel_nodes(new_branch, {sink: branch})
        if branch in self.graph:
            graph = _remove_ancestors(self.graph, branch)
            graph.nodes[branch].clear()
        else:
            graph = self.graph

        intersection_nodes = set(graph.nodes) & set(new_branch.nodes) - {branch}

        for node in intersection_nodes:
            if graph.pred[node] != new_branch.pred[node]:
                raise ValueError(
                    f"Node inputs differ for node '{node}':\n"
                    f"  {graph.pred[node]}\n"
                    f"  {new_branch.pred[node]}\n"
                )
            if graph.nodes[node] != new_branch.nodes[node]:
                raise ValueError(f"Node data differs for node '{node}'")

        graph = nx.compose(graph, new_branch)

        # Delay setting graph until we know no step fails
        self._node_values = self._node_values.merge(other._node_values)
        self.graph = graph



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from collections.abc import Hashable, Mapping, Sequence

import networkx as nx
import numpy as np
import pandas as pd
import pytest
import scipp as sc
import xarray as xr

import cyclebane as cb


def idx(
    name: str, *index: int, offset=None, dims: tuple[str, ...] = ('dim_0', 'dim_1')
) -> cb.graph.NodeName:
    """Helper to create a NodeName with a tuple of indices."""
    return cb.graph.NodeName(
        name,
        cb.graph.IndexValues(dims[offset : len(index) + (offset or 0)], tuple(index)),
    )


@pytest.mark.parametrize('params', [{}, pd.DataFrame()])
def test_map_can_map_over_empty(params) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map(params)
    assert len(mapped.to_networkx().nodes) == 2


@pytest.mark.parametrize(
    'params',
    [
        {'c': [1, 2]},
        {'a': [1, 2], 'c': [1, 2]},
        pd.DataFrame({'a': [1, 2], 'c': [1, 2]}),
    ],
)
def test_map_adds_node_when_mapping_nonexistent_node(params) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map(params)
    result = mapped.to_networkx()
    assert result.nodes[idx('c', 0)] == {'value': 1}
    assert result.nodes[idx('c', 1)] == {'value': 2}


def test_map_raises_if_mapping_non_source_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    with pytest.raises(ValueError, match="Mapped node 'b' is not a source node"):
        graph.map({'b': [1, 2]})


def test_map_raises_if_mapping_previously_mapped_source_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2]})
    with pytest.raises(ValueError, match="Node 'a' has already been mapped"):
        mapped.map({'a': [1, 2]})


def test_map_raises_if_shapes_of_values_are_incompatible() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    with pytest.raises(
        ValueError, match="value sequences in a map operation must have the same shape"
    ):
        graph.map({'a': [1, 2], 'b': [1, 2, 3]})


def test_map_over_list_adds_value_attrs_to_source_nodes() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    result = mapped.to_networkx()

    assert result.nodes[idx('a', 0)] == {'value': 1}
    assert result.nodes[idx('a', 1)] == {'value': 2}
    assert result.nodes[idx('a', 2)] == {'value': 3}
    assert result.nodes[idx('b', 0)] == {}
    assert result.nodes[idx('b', 1)] == {}
    assert result.nodes[idx('b', 2)] == {}


def test_map_does_not_duplicate_unrelated_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('x', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    result = mapped.to_networkx()
    assert len(result.nodes) == 3 + 3 + 1


def test_chained_map_over_list() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.add_edge('x', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]}).map({'x': [4, 5]})
    result = mapped.to_networkx()

    assert result.nodes[idx('a', 0)] == {'value': 1}
    assert result.nodes[idx('a', 1)] == {'value': 2}
    assert result.nodes[idx('a', 2)] == {'value': 3}
    assert result.nodes[idx('x', 0, offset=1)] == {'value': 4}
    assert result.nodes[idx('x', 1, offset=1)] == {'value': 5}


def test_map_does_not_descend_into_nested_lists() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [[1, 2], [3, 4]]})
    assert mapped.index_names == ('dim_0',)
    assert len(mapped.to_networkx().nodes) == 2 + 2


def test_map_adds_axis_in_position_0_like_numpy_stack() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]}).map({'b': [4, 5]})

    reduced = mapped.reduce('c', name='sum', axis=0)
    result = reduced.to_networkx()
    # Axis 0 should have length 2, so reducing it should leave us with 3 sink nodes,
    # i.e., the ones relating to the *first* call to map.
    sink_nodes = [node for node, degree in result.out_degree() if degree == 0]
    assert len(sink_nodes) == 3


def test_map_2d_numpy_array_sets_up_default_index_names() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': np.array([[1, 2, 3], [4, 5, 6]])})
    assert mapped.index_names == ('dim_0', 'dim_1')
    assert len(mapped.to_networkx().nodes) == 3 * 2 * 2


def test_map_2d_xarray_dataarray_uses_dims_as_index_names() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    da = xr.DataArray(dims=('x', 'y'), data=[[1, 2, 3], [4, 5, 6]])
    mapped = graph.map({'a': da})
    assert mapped.index_names == ('x', 'y')
    assert len(mapped.to_networkx().nodes) == 3 * 2 * 2


def test_map_pandas_dataframe() -> None:
    params = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map(params)
    result = mapped.to_networkx()
    assert len(result.nodes) == 3 * 3

    assert result.nodes[idx('a', 0)] == {'value': 1}
    assert result.nodes[idx('a', 1)] == {'value': 2}
    assert result.nodes[idx('a', 2)] == {'value': 3}

    assert result.nodes[idx('b', 0)] == {'value': 4}
    assert result.nodes[idx('b', 1)] == {'value': 5}
    assert result.nodes[idx('b', 2)] == {'value': 6}


def test_map_pandas_dataframe_sets_up_default_index_name() -> None:
    params = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map(params)
    assert mapped.index_names == ('dim_0',)


def test_map_pandas_dataframe_uses_index_name() -> None:
    params = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    params.index.name = 'abcde'
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map(params)
    assert mapped.index_names == ('abcde',)


def test_map_pandas_dataframe_uses_index_values() -> None:
    params = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    params.index = [11, 22, 33]
    params.index.name = 'abcde'
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map(params)
    for node in mapped.to_networkx().nodes:
        assert node.index.axes == ('abcde',)
        assert node.index.values[0] in [11, 22, 33]


def test_map_pandas_dataframe_with_type_as_col_name_works() -> None:
    # This is a slightly obscure way of naming columns, but it is used by Sciline.
    # There is a special test for it since DataFrame.__getitem__ does not work with
    # this kind of column name.
    raw_params = {int: [1, 2, 3], float: [0.1, 0.2, 0.3]}
    params = pd.DataFrame(raw_params)
    g = nx.DiGraph()
    g.add_edge(int, 'a')
    g.add_edge(float, 'a')

    graph = cb.Graph(g)
    mapped = graph.map(params)
    result = mapped.to_networkx()
    assert len(result.nodes) == 3 * 3

    assert result.nodes[idx(int, 0)] == {'value': 1}
    assert result.nodes[idx(int, 1)] == {'value': 2}
    assert result.nodes[idx(int, 2)] == {'value': 3}

    assert result.nodes[idx(float, 0)] == {'value': 0.1}
    assert result.nodes[idx(float, 1)] == {'value': 0.2}
    assert result.nodes[idx(float, 2)] == {'value': 0.3}


def test_map_scipp_variable_uses_dims_as_index_names() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    x = sc.array(dims=['x'], values=[1, 2, 3], unit='m')
    mapped = graph.map({'a': x})
    assert mapped.index_names == ('x',)
    result = mapped.to_networkx()

    a_data = [data for node, data in result.nodes(data=True) if node.name == 'a']
    a_values = sc.concat([data['value'] for data in a_data], 'x')
    assert sc.identical(a_values.data, x)


def test_map_2d_scipp_variable_uses_dims_as_index_names() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    values = sc.array(dims=['x', 'y'], values=[[1, 2, 3], [4, 5, 6]], unit='m')
    mapped = graph.map({'a': values})
    assert mapped.index_names == ('x', 'y')
    result = mapped.to_networkx()

    a_data = [data for node, data in result.nodes(data=True) if node.name == 'a']
    a_values = sc.concat([data['value'] for data in a_data], 'y')
    assert sc.identical(a_values[0:3].data, values['x', 0])
    assert sc.identical(a_values[3:6].data, values['x', 1])


def test_map_scipp_data_array_uses_coord_as_indices_if_present() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    values = sc.array(dims=['x', 'y'], values=[[1, 2, 3], [4, 5, 6]], unit='m')
    da = sc.DataArray(
        values, coords={'y': sc.linspace(dim='y', start=0, stop=1, num=3, unit='mm')}
    )
    mapped = graph.map({'a': da})
    assert mapped.index_names == ('x', 'y')
    result = mapped.to_networkx()

    a_data = [data for node, data in result.nodes(data=True) if node.name == 'a']
    a_values = sc.concat([data['value'] for data in a_data], 'y')
    del a_values.coords['x']  # auto range index not in reference
    assert sc.identical(a_values[0:3], da['x', 0])
    assert sc.identical(a_values[3:6], da['x', 1])

    for name in ['a', 'b']:
        # Note that due to current shortcomings in scipp the indices are tuples of the
        # form (value, unit) and not the corresponding 0-D scipp.Variable.
        assert idx(name, 0, (0.0, 'mm'), dims=('x', 'y')) in result
        assert idx(name, 0, (0.5, 'mm'), dims=('x', 'y')) in result
        assert idx(name, 0, (1.0, 'mm'), dims=('x', 'y')) in result
        assert idx(name, 1, (0.0, 'mm'), dims=('x', 'y')) in result
        assert idx(name, 1, (0.5, 'mm'), dims=('x', 'y')) in result
        assert idx(name, 1, (1.0, 'mm'), dims=('x', 'y')) in result


def test_map_xarray_data_array_uses_coord_as_indices_if_present() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    da = xr.DataArray(
        dims=('x', 'y'),
        data=np.array([[1, 2, 3], [4, 5, 6]]),
        coords={'y': np.linspace(start=0, stop=1, num=3)},
    )
    mapped = graph.map({'a': da})
    assert mapped.index_names == ('x', 'y')
    result = mapped.to_networkx()

    for name in ['a', 'b']:
        assert idx(name, 0, 0.0, dims=('x', 'y')) in result
        assert idx(name, 0, 0.5, dims=('x', 'y')) in result
        assert idx(name, 0, 1.0, dims=('x', 'y')) in result
        assert idx(name, 1, 0.0, dims=('x', 'y')) in result
        assert idx(name, 1, 0.5, dims=('x', 'y')) in result
        assert idx(name, 1, 1.0, dims=('x', 'y')) in result


def test_reduce_scipp_mapped() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    x = sc.array(dims=['x'], values=[1, 2, 3], unit='m')
    mapped = graph.map({'a': x})
    reduced = mapped.reduce('b', name='sum', index='x')

    assert 'sum' in reduced.graph


def test_map_with_previously_mapped_index_name_raises() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    values = sc.arange('x', 3)

    mapped = graph.map({'a': values})
    with pytest.raises(ValueError, match="Conflicting new index names"):
        mapped.map({'b': values})


def test_map_multiple_joint_index() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2], 'b': [4, 5]})
    # a and b have a common index, so we get 2 c's, not 4.
    assert len(mapped.to_networkx().nodes) == 2 + 2 + 2


def test_map_reduce() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.add_edge('x', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]}).map({'x': [4, 5]})
    reduced = mapped.reduce(name='func', axis=1)
    # Axis 0 reduces 'x', so there are 2 reduce nodes.
    assert len(reduced.to_networkx().nodes) == 19
    # Axis 1 reduces 'a', so there are 3 reduce nodes.
    reduced = mapped.reduce(name='func', axis=0)
    result = reduced.to_networkx()
    assert len(result.nodes) == 20

    assert result.nodes[idx('a', 0)] == {'value': 1}
    assert result.nodes[idx('a', 1)] == {'value': 2}
    assert result.nodes[idx('a', 2)] == {'value': 3}
    assert result.nodes[idx('x', 0, offset=1)] == {'value': 4}
    assert result.nodes[idx('x', 1, offset=1)] == {'value': 5}


def test_reduce_all_axes() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]}).map({'b': [4, 5]})
    reduced = mapped.reduce(name='sum', attrs={'func': 'sum'})
    # No axis or index given, all axes are reduced, so the new node has no index part.
    assert 'sum' in reduced.graph
    assert reduced.graph.nodes['sum'] == {'func': 'sum'}


def test_reduce_preserves_reduced_index_names() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    graph = cb.Graph(g).map({'a': sc.ones(dims=['x', 'y'], shape=(2, 3))})
    reduced = graph.reduce('b', name='combine')
    # The new node is reduced, but the graph in its entirety still has the dims.
    assert reduced.index_names == ('x', 'y')


def test_reduce_raises_if_new_node_name_exists() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('other', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    with pytest.raises(ValueError, match="Node 'other' already exists in the graph."):
        mapped.reduce(name='other')


@pytest.mark.parametrize('indexer', [{'axis': 1}, {'index': 'y'}])
def test_reduce_raises_if_axis_or_does_not_exist(indexer) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': sc.arange('x', 3)})
    with pytest.raises(ValueError, match="does not have"):
        mapped.reduce(name='combine', **indexer)


@pytest.mark.parametrize('indexer', [{'axis': 1}, {'index': 'y'}])
def test_reduce_raises_if_node_does_not_have_the_specified_axis_or_index(
    indexer,
) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': sc.arange('x', 3)})
    with pytest.raises(ValueError, match="Node 'b' does not have "):
        # We mapped 'a' but 'b' is not a descendant of 'a'.
        mapped.reduce('b', name='combine', **indexer)


def test_reduce_works_with_related_unmapped_nodes() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    reduced = mapped.reduce('c', name='combine')
    result = reduced.to_networkx()
    assert len(result.nodes) == 3 + 3 + 1 + 1  # 3a + 3b + 1c + 1reduce


@pytest.mark.parametrize('indexer', [{'axis': 0}, {'index': 'x'}])
def test_can_reduce_same_axis_or_index_on_multiple_nodes(indexer) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('a', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': sc.ones(dims=['x'], shape=(3,))})
    reduced = (
        mapped.reduce('a', name='reduce-a', **indexer)
        .reduce('b', name='reduce-b', **indexer)
        .reduce('c', name='reduce-c', **indexer)
    )
    result = reduced.to_networkx()
    assert len(result.nodes) == 3 + 3 + 3 + 1 + 1 + 1


def test_can_reduce_same_node_multiple_times() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': sc.ones(dims=['x'], shape=(3,))})
    reduced = mapped.reduce('b', name='c1', axis=0).reduce('b', name='c2', axis=0)
    result = reduced.to_networkx()
    assert len(result.nodes) == 3 + 3 + 1 + 1
    c1_parents = list(result.predecessors('c1'))
    c2_parents = list(result.predecessors('c2'))
    assert c1_parents == c2_parents


def test_can_reduce_different_axes_or_indices_of_same_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': sc.ones(dims=['x', 'y'], shape=(3, 3))})
    reduced = (
        mapped.reduce('b', name='c0', axis=0)
        .reduce('b', name='c1', axis=1)
        .reduce('b', name='cx', index='x')
        .reduce('b', name='cy', index='y')
    )
    # Helper so we can get all the parents of the reduce nodes.
    helper = (
        reduced.reduce('c0', name='d0', axis=0)
        .reduce('c1', name='d1', axis=0)
        .reduce('cx', name='dx', index='y')
        .reduce('cy', name='dy', index='x')
    ).to_networkx()
    reduced = reduced.to_networkx()

    assert len(reduced.nodes) == 9 + 9 + 4 * 3
    c0s = list(helper.predecessors('d0'))
    c1s = list(helper.predecessors('d1'))
    cxs = list(helper.predecessors('dx'))
    cys = list(helper.predecessors('dy'))

    for c0, cx in zip(c0s, cxs, strict=True):
        c0_parents = list(reduced.predecessors(c0))
        cx_parents = list(reduced.predecessors(cx))
        assert c0_parents == cx_parents

    for c1, cy in zip(c1s, cys, strict=True):
        c1_parents = list(reduced.predecessors(c1))
        cy_parents = list(reduced.predecessors(cy))
        assert c1_parents == cy_parents


def test_axis_in_reduce_refers_to_node_axis_not_graph_axis() -> None:
    # TODO Is this actually the behavior we want?
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    graph = graph.map({'a': sc.ones(dims=['x', 'y', 'z'], shape=(2, 2, 2))})
    graph = graph.reduce('b', name='c', index='x')

    # Axis 1 of the graph is 'y', but we are reducing at 'c' which is (y, z)
    # so axis 1 of the reduce is 'z'.
    result = graph.reduce('c', name='d', axis=1).to_networkx()
    d_nodes = [n for n in result.nodes if n.name == 'd']
    # 'y' is left, even though axis 1 of the graph is 'y'.
    assert all(n.index.axes == ('y',) for n in d_nodes)


def test_delitem_removes_ancestors_and_data_but_keeps_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.add_node('b', value='x')

    graph = cb.Graph(g)
    del graph['b']
    result = graph.to_networkx()
    assert list(result.nodes) == ['b', 'c']
    assert result.nodes['b'] == {}


def test_delitem_preserves_ancestors_with_other_path_to_graph() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.add_node('b', value='x')
    # a has another path to c so it will we kept
    g.add_edge('a', 'c')

    graph = cb.Graph(g)
    del graph['b']
    result = graph.to_networkx()
    assert list(result.nodes) == ['a', 'b', 'c']
    assert result.nodes['b'] == {}
    assert not result.has_edge('a', 'b')


def test_delitem_raises_when_removing_mapped_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.add_node('b', value='x')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    with pytest.raises(ValueError, match="Cannot delete mapped node."):
        del mapped['b']


def test_delitem_can_remove_reduced_node_depending_on_mapped_nodes() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('x', 'y')  # unrelated
    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]}).reduce('b', name='c')

    del mapped['c']

    result = mapped.to_networkx()
    assert set(result.nodes) == {'x', 'y', 'c'}
    assert result.nodes['c'] == {}


def test_setitem_raises_TypeError_if_given_networkx_graph() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    with pytest.raises(TypeError):
        graph['a'] = nx.DiGraph()


def test_setitem_with_other_graph_keeps_nodename_of_key_but_replaces_node_data() -> (
    None
):
    g1 = nx.DiGraph()
    g1.add_edge('b', 'a')
    g1.nodes['b']['attr'] = 1
    g2 = nx.DiGraph()
    g2.add_edge('d', 'c')
    g2.nodes['c']['attr'] = 2

    graph = cb.Graph(g1)
    graph['b'] = cb.Graph(g2)
    assert 'b' in graph.to_networkx()
    nx_graph = graph.to_networkx()
    assert set(nx_graph.nodes) == {'a', 'b', 'd'}
    assert len(nx_graph.edges) == 2
    assert nx_graph.has_edge('d', 'b')
    assert nx_graph.has_edge('b', 'a')
    assert nx_graph.nodes['b'] == {'attr': 2}


def test_setitem_raises_on_conflicting_ancestor_node_data() -> None:
    g1 = nx.DiGraph()
    g1.add_edge('a', 'b')
    g1.nodes['a']['attr'] = 1
    g1.add_edge('x', 'b')
    g2 = nx.DiGraph()
    g2.add_edge('a', 'x')
    g2.nodes['a']['attr'] = 2

    graph = cb.Graph(g1)
    with pytest.raises(ValueError, match="Node data differs for node 'a'"):
        graph['x'] = cb.Graph(g2)


def test_setitem_raises_on_conflicting_input_nodes_in_ancestor() -> None:
    g1 = nx.DiGraph()
    g1.add_edge('a1', 'b')
    g1.add_edge('b', 'c')
    g1.add_edge('x', 'c')
    g2 = nx.DiGraph()
    g2.add_edge('a2', 'b')
    g2.add_edge('b', 'x')

    graph = cb.Graph(g1)
    with pytest.raises(ValueError, match="Node inputs differ for node 'b'"):
        graph['x'] = cb.Graph(g2)


def test_setitem_replaces_nodes_that_are_not_ancestors_of_unrelated_node() -> None:
    g1 = nx.DiGraph()
    g1.add_edge('a', 'b')
    g1.add_edge('b', 'c')
    g1.add_edge('c', 'd')
    graph = cb.Graph(g1)
    g2 = nx.DiGraph()
    g2.add_edge('b', 'c')
    graph['c'] = cb.Graph(g2)
    assert 'a' not in graph.to_networkx()


def test_setitem_preserves_nodes_that_are_ancestors_of_unrelated_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.add_edge('b', 'd')
    g.add_edge('c', 'd')
    graph = cb.Graph(g)
    graph['c'] = graph['c']
    nx.utils.graphs_equal(graph.to_networkx(), g)


def test_getitem_returns_graph_containing_only_key_and_ancestors() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.add_edge('c', 'd')
    g.add_edge('x', 'd')

    graph = cb.Graph(g)
    subgraph = graph['c']
    result = subgraph.to_networkx()
    assert len(result.nodes) == 3
    assert len(result.edges) == 2
    assert 'a' in result
    assert 'b' in result
    assert 'c' in result


def test_getitem_setitem_with_no_effects() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.nodes['a']['value'] = 1

    graph = cb.Graph(g)
    graph['b'] = graph['b']
    assert graph.to_networkx().nodes['a']['value'] == 1
    # Note that we cannot currently do the following:
    #   mapped = graph.map({'a': [1, 2, 3]})
    #   mapped['b'] = mapped['b']
    # because the check for compatible indices/node-values is not implemented.


def test_getitem_keeps_only_relevant_indices() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    assert mapped['a'].indices == {'dim_0': range(3)}
    assert mapped['b'].indices == {}  # Not mapped
    assert mapped['c'].indices == {'dim_0': range(3)}


def test_getitem_keeps_only_relevant_node_values() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    # This fails due to existing mapping...
    with pytest.raises(ValueError, match='has already been mapped'):
        mapped.map({'a': [1, 2]})
    # ... but getitem drops the 'a' mapping, so we can map 'a' again:
    mapped['b'].map({'a': [1, 2]})


def test_getitem_on_mapped_graph_with_base_node_name_returns_mapped_node() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    result = mapped['b'].to_networkx()
    assert len(result.nodes) == 6


def test_setitem_with_mapped_sink_node_raises_if_target_is_not_mapped() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    b = cb.Graph(nx.DiGraph()).map({'b': [11, 12]})
    with pytest.raises(
        NotImplementedError, match="Mapped nodes not supported yet in __setitem__"
    ):
        graph['b'] = b


def test_setitem_with_mapped_operands_raises_on_conflict() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    d = cb.Graph(nx.DiGraph()).map({'b': [11, 12]}).reduce('b', name='d')
    with pytest.raises(ValueError, match="Conflicting new index names"):
        mapped['x'] = d


def test_setitem_currently_does_not_allow_compatible_indices() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('c', 'd')

    graph = cb.Graph(g)
    mapped1 = graph.map({'a': [1, 2, 3]})
    mapped2 = graph['d'].map({'c': [11, 12, 13]}).reduce('d', name='e')
    # Note: This is a limitation of the current implementation. We could check if the
    # indices are identical and allow this. For simplicity we currently do not.
    with pytest.raises(ValueError, match="Conflicting new index names"):
        mapped1['x'] = mapped2


@pytest.mark.parametrize(
    'node_values',
    [
        {'a': [1, 2, 3]},
        {'a': [1, 2, 3], 'b': [11, 12, 13]},
        {'a': np.array([1, 2, 3])},
        {'a': np.array([1, 2, 3]), 'b': np.array([11, 12, 13])},
        pd.DataFrame({'a': [1, 2, 3]}),
        pd.DataFrame({'a': [1, 2, 3], 'b': [11, 12, 13]}),
        {'a': sc.array(dims=['x'], values=[1, 2, 3])},
        {
            'a': sc.array(dims=['x'], values=[1, 2, 3]),
            'b': sc.array(dims=['x'], values=[11, 12, 13]),
        },
        {'a': xr.DataArray(dims=('x',), data=[1, 2, 3])},
        {
            'a': xr.DataArray(dims=('x',), data=[1, 2, 3]),
            'b': xr.DataArray(dims=('x',), data=[11, 12, 13]),
        },
    ],
)
def test_setitem_allows_compatible_node_values(node_values) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'c')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    mapped = graph.map(node_values).reduce('c', name='d')
    mapped['x'] = mapped['d']
    assert len(mapped.index_names) == 1


def test_setitem_raises_if_node_values_equivalent_but_of_different_type() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    graph = cb.Graph(g)
    mapped1 = graph.map({'a': [1, 2]}).reduce('b', name='d')
    mapped2 = graph.map({'a': np.array([1, 2])}).reduce('b', name='d')
    # One could imagine treating this as equivalent, but we are strict in the
    # comparison.
    with pytest.raises(ValueError, match="Conflicting new index names"):
        mapped1['x'] = mapped2['d']


def test_setitem_raises_if_node_values_incompatible() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    graph = cb.Graph(g)
    mapped1 = graph.map({'a': [1, 2]}).reduce('b', name='d')
    mapped2 = graph.map({'a': sc.array(dims=('x',), values=[1, 2])}).reduce(
        'b', name='d'
    )
    with pytest.raises(ValueError, match="has already been mapped"):
        mapped1['x'] = mapped2['d']


def test_setitem_does_currently_not_support_slice_assignment() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')

    graph = cb.Graph(g)
    with pytest.raises(NotImplementedError):
        graph['b':'b'] = graph['b']
    with pytest.raises(NotImplementedError):
        graph['b':'a'] = graph['b']


def test_setitem_raises_if_value_graph_does_not_have_unique_sink() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('a', 'c')

    graph = cb.Graph(g)
    with pytest.raises(ValueError, match="Graph must have exactly one sink node"):
        graph['a'] = graph


@pytest.mark.parametrize(
    'param_table',
    [{'a': [1, 2, 3]}, {'a': np.array([1, 2, 3])}, pd.DataFrame({'a': [1, 2, 3]})],
)
def test_slice_by_position(param_table: Mapping[Hashable, Sequence[int]]) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map(param_table)
    sliced = mapped.by_position('dim_0')[1:3]
    result = sliced.to_networkx()
    assert cb.graph.NodeName('a', cb.graph.IndexValues(('dim_0',), (0,))) not in result
    assert cb.graph.NodeName('a', cb.graph.IndexValues(('dim_0',), (1,))) in result
    assert cb.graph.NodeName('a', cb.graph.IndexValues(('dim_0',), (2,))) in result

    assert idx('a', 0) not in result
    assert result.nodes[idx('a', 1)] == {'value': 2}
    assert result.nodes[idx('a', 2)] == {'value': 3}


@pytest.mark.parametrize(
    'values',
    [
        np.array([[1, 2, 3], [4, 5, 6]]),
        xr.DataArray(dims=('dim_0', 'dim_1'), data=[[1, 2, 3], [4, 5, 6]]),
    ],
)
def test_by_position_2d_slice_outer(values) -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': values})
    sliced = mapped.by_position('dim_0')[1:]
    result = sliced.to_networkx()

    assert idx('a', 0, 0) not in result
    assert idx('a', 0, 1) not in result
    assert idx('a', 0, 2) not in result
    assert idx('a', 1, 0) in result
    assert idx('a', 1, 1) in result
    assert idx('a', 1, 2) in result

    a_data = [data for node, data in result.nodes(data=True) if node.name == 'a']
    a_values = [data['value'] for data in a_data]
    assert a_values[0:3] == [4, 5, 6]


def test_by_position_2d_slice_inner() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')

    graph = cb.Graph(g)
    mapped = graph.map({'a': np.array([[1, 2, 3], [4, 5, 6]])})
    sliced = mapped.by_position('dim_1')[:2]
    result = sliced.to_networkx()

    assert idx('a', 0, 0) in result
    assert idx('a', 0, 1) in result
    assert idx('a', 0, 2) not in result
    assert idx('a', 1, 0) in result
    assert idx('a', 1, 1) in result
    assert idx('a', 1, 2) not in result

    assert result.nodes[idx('a', 0, 0)] == {'value': 1}
    assert result.nodes[idx('a', 0, 1)] == {'value': 2}
    assert result.nodes[idx('a', 1, 0)] == {'value': 4}
    assert result.nodes[idx('a', 1, 1)] == {'value': 5}


def test_node_attrs_are_preserved() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.nodes['a']['attr'] = 1

    graph = cb.Graph(g)
    result = graph.to_networkx()
    assert result.nodes['a'] == {'attr': 1}


def test_node_attrs_are_preserved_in_getitem() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.add_edge('b', 'c')
    g.nodes['a']['attr1'] = 1
    g.nodes['b']['attr2'] = 2
    g.nodes['c']['attr3'] = 3

    graph = cb.Graph(g)
    result = graph['c'].to_networkx()
    assert result.nodes['a'] == {'attr1': 1}
    assert result.nodes['b'] == {'attr2': 2}


def test_node_attrs_are_preserved_in_setitem() -> None:
    g1 = nx.DiGraph()
    g1.add_edge('a', 'c')
    g1.add_edge('b', 'c')
    g1.nodes['a']['attr1'] = 1
    g1.nodes['b']['attr2'] = 2
    g1.nodes['c']['attr3'] = 3

    graph = cb.Graph(g1)
    g2 = nx.DiGraph()
    g2.add_edge('x', 'b')
    g2.nodes['x']['attr4'] = 4
    g2.nodes['b']['attr5'] = 5

    graph['b'] = cb.Graph(g2)

    result = graph.to_networkx()
    assert result.nodes['a'] == {'attr1': 1}
    assert result.nodes['b'] == {'attr5': 5}  # b was replaced
    assert result.nodes['c'] == {'attr3': 3}
    assert result.nodes['x'] == {'attr4': 4}


def test_node_attrs_are_preserved_in_map() -> None:
    g = nx.DiGraph()
    g.add_edge('a', 'b')
    g.nodes['a']['attr'] = 11
    g.nodes['b']['attr'] = 22

    graph = cb.Graph(g)
    mapped = graph.map({'a': [1, 2, 3]})
    value_attr = 'myvalue'
    result = mapped.to_networkx(value_attr=value_attr)

    assert result.nodes[idx('a', 0)] == {'attr': 11, 'myvalue': 1}
    assert result.nodes[idx('a', 1)] == {'attr': 11, 'myvalue': 2}
    assert result.nodes[idx('a', 2)] == {'attr': 11, 'myvalue': 3}
    assert result.nodes[idx('b', 0)] == {'attr': 22}
    assert result.nodes[idx('b', 1)] == {'attr': 22}
    assert result.nodes[idx('b', 2)] == {'attr': 22}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

from .graph import Graph

__all__ = ["Graph"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Hashable, Iterable, Iterator, Mapping, Sequence
from types import ModuleType
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

if TYPE_CHECKING:
    import numpy
    import pandas
    import scipp
    import xarray

IndexName = Hashable
IndexValue = Hashable

T = TypeVar('T', bound='ValueArray')


class ValueArray(ABC):
    """
    Abstract base class for a series of values with an index that can be sliced.

    Used by :py:class:`NodeValues` to store the values of a given node in a graph. The
    abstraction allows for the use of different data structures to store the values of
    nodes in a graph, such as pandas.DataFrame, xarray.DataArray, numpy.ndarray, or
    simple Python iterables.
    """

    _registry: ClassVar = []

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        ValueArray._registry.append(cls)

    @staticmethod
    def from_array_like(values: Any, *, axis_zero: int = 0) -> ValueArray:
        # Reversed to ensure SequenceAdapter is tried last, as it is the most general
        # SequenceAdapter is defined right after this class so it is registered first
        for subclass in reversed(ValueArray._registry):
            if (a := subclass.try_from(values, axis_zero=axis_zero)) is not None:
                return a
        raise ValueError(f'Cannot create ValueArray from {values}')

    @staticmethod
    @abstractmethod
    def try_from(obj: Any, *, axis_zero: int = 0) -> ValueArray | None: ...

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other):
            return NotImplemented
        return self._equal(other)

    def __ne__(self, other: object) -> bool:
        return not self == other

    @abstractmethod
    def _equal(self: T, other: T) -> bool: ...

    @abstractmethod
    def sel(self, key: tuple[tuple[IndexName, IndexValue], ...]) -> Any:
        """Return data by selecting from index with given name and index value."""

    def loc(self, key: dict[IndexName, slice]) -> ValueArray:
        if not all(isinstance(i, slice) for i in key.values()):
            raise ValueError('ValueArray.loc only accepts slices, not integers')
        if not set(key).issubset(set(self.index_names)):
            raise ValueError(
                f'ValueArray.loc got {key.keys()}, not a subset of {self.index_names}'
            )
        return self[key]

    @abstractmethod
    def __getitem__(self, key: dict[IndexName, slice]) -> ValueArray:
        pass

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        pass

    @property
    @abstractmethod
    def index_names(self) -> tuple[IndexName, ...]:
        pass

    @property
    @abstractmethod
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        pass


class SequenceAdapter(ValueArray):
    def __init__(
        self,
        values: Sequence[Any],
        *,
        index: Iterable[IndexValue] | None = None,
        axis_zero: int = 0,
    ):
        self._values = values
        self._index = index or range(len(values))
        self._axis_zero = axis_zero

    @staticmethod
    def try_from(obj: Any, *, axis_zero: int = 0) -> SequenceAdapter | None:
        return SequenceAdapter(obj, axis_zero=axis_zero)

    def _equal(self, other: SequenceAdapter) -> bool:
        return (
            self._values == other._values
            and self._index == other._index
            and self._axis_zero == other._axis_zero
        )

    def sel(self, key: tuple[tuple[IndexName, IndexValue], ...]) -> Any:
        if len(key) != 1:
            raise ValueError('SequenceAdapter only supports single index')
        _, i = key[0]
        return self._values[self._index.index(i)]

    def __getitem__(self, key: dict[IndexName, slice]) -> SequenceAdapter:
        _, i = next(iter(key.items()))
        return SequenceAdapter(
            self._values[i], index=self._index[i], axis_zero=self._axis_zero
        )

    @property
    def shape(self) -> tuple[int, ...]:
        return (len(self._values),)

    @property
    def index_names(self) -> tuple[IndexName, ...]:
        return (f'dim_{self._axis_zero}',)

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        return {f'dim_{self._axis_zero}': self._index}


class PandasSeriesAdapter(ValueArray):
    def __init__(self, series: pandas.Series, *, axis_zero: int = 0):
        self._series = series
        self._axis_zero = axis_zero

    @staticmethod
    def try_from(obj: Any, *, axis_zero: int = 0) -> PandasSeriesAdapter | None:
        try:
            import pandas
        except ModuleNotFoundError:
            return None
        if isinstance(obj, pandas.Series):
            return PandasSeriesAdapter(obj, axis_zero=axis_zero)

    def _equal(self, other: PandasSeriesAdapter) -> bool:
        return (
            self._series.equals(other._series) and self._axis_zero == other._axis_zero
        )

    def sel(self, key: tuple[tuple[IndexName, IndexValue], ...]) -> Any:
        if len(key) != 1:
            raise ValueError('PandasSeriesAdapter only supports single index')
        index_name, i = key[0]
        if index_name != self.index_names[0]:
            raise ValueError(
                f'Unexpected index name {index_name} for PandasSeriesAdapter with '
                f'index names {self.index_names}'
            )
        return self._series.loc[i]

    def __getitem__(self, key: dict[IndexName, slice]) -> PandasSeriesAdapter:
        _, i = next(iter(key.items()))
        return PandasSeriesAdapter(self._series[i], axis_zero=self._axis_zero)

    @property
    def shape(self) -> tuple[int, ...]:
        return (len(self._series),)

    @property
    def index_names(self) -> tuple[IndexName, ...]:
        index_name = (
            self._series.index.name
            if self._series.index.name is not None
            else f'dim_{self._axis_zero}'
        )
        return (index_name,)

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        return {self.index_names[0]: self._series.index}


class XarrayDataArrayAdapter(ValueArray):
    def __init__(
        self,
        data_array: xarray.DataArray,
    ):
        default_indices = {
            dim: range(size)
            for dim, size in data_array.sizes.items()
            if dim not in data_array.coords
        }
        self._data_array = data_array.assign_coords(default_indices)

    @staticmethod
    def try_from(obj: Any, *, axis_zero: int = 0) -> XarrayDataArrayAdapter | None:
        try:
            import xarray

            if isinstance(obj, xarray.DataArray):
                return XarrayDataArrayAdapter(obj)
        except ModuleNotFoundError:
            pass

    def _equal(self, other: XarrayDataArrayAdapter) -> bool:
        return self._data_array.identical(other._data_array)

    def sel(self, key: tuple[tuple[IndexName, IndexValue], ...]) -> Any:
        return self._data_array.sel(dict(key))

    def __getitem__(self, key: dict[IndexName, slice]) -> XarrayDataArrayAdapter:
        return XarrayDataArrayAdapter(self._data_array.isel(key))

    @property
    def shape(self) -> tuple[int, ...]:
        return self._data_array.shape

    @property
    def index_names(self) -> tuple[IndexName, ...]:
        return tuple(self._data_array.dims)

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        return {
            dim: self._data_array.coords[dim].values for dim in self._data_array.dims
        }


class ScippDataArrayAdapter(ValueArray):
    def __init__(self, data_array: scipp.DataArray, scipp: ModuleType):
        default_indices = {
            dim: scipp.arange(dim, size, unit=None)
            for dim, size in data_array.sizes.items()
            if dim not in data_array.coords
        }
        self._data_array = data_array.assign_coords(default_indices)
        self._scipp = scipp

    @staticmethod
    def try_from(obj: Any, *, axis_zero: int = 0) -> ScippDataArrayAdapter | None:
        try:
            import scipp

            if isinstance(obj, scipp.Variable):
                return ScippDataArrayAdapter(scipp.DataArray(obj), scipp=scipp)
            if isinstance(obj, scipp.DataArray):
                return ScippDataArrayAdapter(obj, scipp=scipp)
        except ModuleNotFoundError:
            pass

    def _equal(self, other: ScippDataArrayAdapter) -> bool:
        return self._scipp.identical(self._data_array, other._data_array)

    def sel(self, key: tuple[tuple[IndexName, IndexValue], ...]) -> Any:
        values = self._data_array
        for dim, value in key:
            # Reconstruct label, to use label-based indexing instead of positional
            if isinstance(value, tuple):
                value, unit = value
            else:
                unit = None
            label = self._scipp.scalar(value, unit=unit)
            # Scipp indexing uses a comma to separate dimension label from the index,
            # unlike Numpy and other libraries where it separates the indices for
            # different axes.
            values = values[dim, label]
        return values

    def __getitem__(self, key: dict[IndexName, slice]) -> ScippDataArrayAdapter:
        values = self._data_array
        for dim, i in key:
            values = values[dim, i]
        return ScippDataArrayAdapter(values, scipp=self._scipp)

    @property
    def shape(self) -> tuple[int, ...]:
        return self._data_array.shape

    @property
    def index_names(self) -> tuple[IndexName, ...]:
        return tuple(self._data_array.dims)

    def _index_for_dim(self, dim: str) -> list[tuple[Any, scipp.Unit]]:
        # Work around some NetworkX errors. Probably scipp.Variable lacks functionality.
        # For now we return a list of tuples, where the first element is the value and
        # the second is the unit.
        coord = self._data_array.coords[dim]
        unit = coord.unit
        if unit is None:
            return coord.values
        unit = str(unit)
        return [(value, unit) for value in coord.values]

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        return {dim: self._index_for_dim(dim) for dim in self._data_array.dims}


class NumpyArrayAdapter(ValueArray):
    def __init__(
        self,
        array: numpy.ndarray,
        *,
        indices: dict[IndexName, Iterable[IndexValue]] | None = None,
        axis_zero: int = 0,
    ):
        import numpy as np

        self._array = np.asarray(array)
        if indices is None:
            indices = {
                f'dim_{i+axis_zero}': range(size)
                for i, size in enumerate(self._array.shape)
            }
        self._indices = indices
        self._axis_zero = axis_zero

    @staticmethod
    def try_from(obj: Any, *, axis_zero: int = 0) -> NumpyArrayAdapter | None:
        try:
            import numpy
        except ModuleNotFoundError:
            return None
        if isinstance(obj, numpy.ndarray):
            return NumpyArrayAdapter(obj, axis_zero=axis_zero)

    def _equal(self, other: NumpyArrayAdapter) -> bool:
        return (
            (self._array == other._array).all()
            and self._indices == other._indices
            and self._axis_zero == other._axis_zero
        )

    def sel(self, key: tuple[tuple[IndexName, IndexValue], ...]) -> Any:
        index_tuple = tuple(self._indices[k].index(i) for k, i in key)
        return self._array[index_tuple]

    def __getitem__(self, key: dict[IndexName, slice]) -> NumpyArrayAdapter:
        return NumpyArrayAdapter(
            self._array[tuple(key.get(k, slice(None)) for k in self._indices)],
            indices={
                index_name: (index_values[key.get(index_name, slice(None))])
                for index_name, index_values in self._indices.items()
            },
            axis_zero=self._axis_zero,
        )

    @property
    def shape(self) -> tuple[int, ...]:
        return self._array.shape

    @property
    def index_names(self) -> tuple[IndexName, ...]:
        return tuple(self._indices)

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        return self._indices


class NodeValues(Mapping[Hashable, ValueArray]):
    """
    A collection of pandas.DataFrame-like objects with distinct indices.

    This is used by :py:class:`Graph` to store the values of nodes in a graph.
    """

    def __init__(self, values: Mapping[Hashable, ValueArray]):
        self._values = values

    def __len__(self) -> int:
        """Return the number of columns."""
        return len(self._values)

    def __iter__(self) -> Iterator[Hashable]:
        """Iterate over the column names."""
        return iter(self._values)

    def __getitem__(self, key: Hashable) -> ValueArray:
        """Return the column with the given name."""
        return self._values[key]

    @staticmethod
    def from_mapping(
        values: Mapping[Hashable, Sequence[Any]], axis_zero: int
    ) -> NodeValues:
        """Construct from a mapping of node names to value sequences."""
        value_arrays = {
            key: ValueArray.from_array_like(value, axis_zero=axis_zero)
            for key, value in values.items()
        }
        shapes = {array.shape for array in value_arrays.values()}
        if len(shapes) > 1:
            raise ValueError(
                'All value sequences in a map operation must have the same shape. '
                'Use multiple map operations if necessary.'
            )
        return NodeValues(value_arrays)

    def merge(self, value_arrays: Mapping[Hashable, ValueArray]) -> NodeValues:
        value_arrays = {
            key: value
            for key, value in value_arrays.items()
            if self.get(key, None) != value
        }
        if value_arrays:
            named = next(iter(value_arrays.values())).index_names
            if any(name in self.indices for name in named):
                raise ValueError(
                    f'Conflicting new index names {named} with existing '
                    f'{tuple(self.indices)}'
                )
        for node in value_arrays:
            if node in self:
                raise ValueError(f"Node '{node}' has already been mapped")
        return NodeValues({**self._values, **value_arrays})

    def get_columns(self, keys: list[Hashable]) -> NodeValues:
        """Select a subset of columns."""
        return NodeValues({key: self._values[key] for key in keys})

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        """Return the indices of the NodeValues object."""
        value_indices = [value.indices for value in self._values.values()]
        return {
            name: index for indices in value_indices for name, index in indices.items()
        }



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import cyclebane as pkg


def test_has_version() -> None:
    assert hasattr(pkg, '__version__')
